/*
Firmware to manage the 'hidden' functions of a RTL8366CB
GBit switch.
(C) 2010 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#define F_CPU 6000000
#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <avr/interrupt.h>
#include "stdout.h"
#include "io.h"
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/eeprom.h>
#include "menu.h"

#define RTL_GENCTL 0

#define RTL_PORTSTATUS_BASE 		0x14
#define RTL_PORTSTATUS_LINKMASK		0x10
#define RTL_PORTSTATUS_SPEEDMASK	0x3
#define RTL_PORTSTATUS_DUPLEXMASK	0x4

#define RTL_VLANTABLE_READBASE	0x18c
#define RTL_VLANTABLE_WRITEBASE	0x185
#define RTL_VLANTABLE_ACCESSCTL	0x180
#define RTL_VLANTABLE_DO_READ	0xE01
#define RTL_VLANTABLE_DO_WRITE	0xF01

#define RTL_VLANCTL_BASE	0x063 //Set to unity on vlan enable: port1=member0, port2=member1 etc
#define RTL_MEMCONF_BASE	0x20 //Member config



unsigned int getPortVid(int port) {
	return rtlRead(RTL_MEMCONF_BASE+(3*port));
}

void setPortVid(int port, int vid) {
	rtlWrite(RTL_MEMCONF_BASE+(3*port), vid);
}

/* Membership is... weird.
Port status		output
T				Tagged as vlan x
T+M				Untagged
M				None
-				None
*/
#define VLAN_NONE 0
#define VLAN_TAGGED 1
#define VLAN_UNTAGGED 2
char setVlanMembership(int port, char type, int vid) {
	unsigned int r1, r2, r3;
	unsigned int mask=(1<<port)|(1<<(port+8));
	rtlWrite(RTL_VLANTABLE_WRITEBASE, vid);
	rtlWrite(RTL_VLANTABLE_ACCESSCTL, RTL_VLANTABLE_DO_READ);
	r1=rtlRead(RTL_VLANTABLE_READBASE);
	r2=rtlRead(RTL_VLANTABLE_READBASE+1);
	r3=rtlRead(RTL_VLANTABLE_READBASE+2);
	r2&=~((1<<port)|(1<<(port+8)));
	if (type!=VLAN_NONE) r2|=(1<<port);
	if (type==VLAN_UNTAGGED) r2|=(1<<(port+8));
	rtlWrite(RTL_VLANTABLE_WRITEBASE, r1);
	rtlWrite(RTL_VLANTABLE_WRITEBASE+1, r2);
	rtlWrite(RTL_VLANTABLE_WRITEBASE+2, r3);
	rtlWrite(RTL_VLANTABLE_ACCESSCTL, RTL_VLANTABLE_DO_WRITE);
	return r2&mask?1:0;
}

void dumpVlanInfo(void) {
	printf_P(PSTR("Untagged incoming packets will get the following VID tag:\n"));
	for (int x=0; x<5; x++) {
		printf_P(PSTR("Port %i: VID %i\n"), x+1, getPortVid(x));
	}
	printf_P(PSTR("Vlans defined and ports in them: (U=Untagged, T=Tagged) (takes a while)\n"));
	printf_P(PSTR("VID   p1 p2 p3 p4 p5\n"));
	for (unsigned int vid=0; vid<4096; vid++) {
		unsigned int r1, r2, r3;
		rtlWrite(RTL_VLANTABLE_WRITEBASE, vid);
		rtlWrite(RTL_VLANTABLE_ACCESSCTL, RTL_VLANTABLE_DO_READ);
		r1=rtlRead(RTL_VLANTABLE_READBASE);
		r2=rtlRead(RTL_VLANTABLE_READBASE+1);
		r3=rtlRead(RTL_VLANTABLE_READBASE+2);
		if (r2!=0) {
			printf_P(PSTR("%04i: "), vid);
			for (int port=0; port<5; port++) {
				int mask=(1<<port);
				if (r2&mask) {
					if ((r2>>8)&mask) putchar('U'); else putchar('T');
				} else {
					putchar(' '); 
				}
				putchar(' '); putchar(' ');
			}
			printf_P(PSTR("\n"), vid);
		}
	}
}


int currentPort;

int gethex(void) {
	char c;
	char first=1;
	int amt=0;
	while (1) {
		c=getchar();
		if (c>='0' && c<='9') {
			amt<<=4;
			amt+=c-'0';
		} else if (c>='A' && c<='F') {
			amt<<=4;
			amt+=c-'A'+10;	
		} else if (c>='a' && c<='f') {
			amt<<=4;
			amt+=c-'a'+10;
		} else {
			if (first) return -1;
			putchar('\n');
			return amt;
		}
		putchar(c);
		first=0;
	}
}


int getdec(void) {
	char c;
	char first=1;
	int amt=0;
	while (1) {
		c=getchar();
		if (c>='0' && c<='9') {
			amt=amt*10;
			amt+=c-'0';
		} else {
			if (first) return -1;
			putchar('\n');
			return amt;
		}
		putchar(c);
		first=0;
	}
}

void waitkey(void) {
	printf_P(PSTR("Press any key to continue.\n"));
	getchar();
}

void dumpEeprom(void) {
	int x,y;
	for (y=0; y<256; y+=16) {
		printf_P(PSTR("%02x: "), y);
		for (x=0; x<16; x++) {
			int i=eepromRead(y+x);
			printf_P(PSTR("%02x "), i);
		}
		printf_P(PSTR("\n"));
	}
	waitkey();
}

void dumpRtlRegisters(void) {
	int x,y;
	for (y=0; y<0x300; y+=8) {
//		if (y==0x80) y=0x100;
		printf_P(PSTR("%03x: "), y);
		for (x=0; x<8; x++) {
			unsigned int i=rtlRead(y+x);
			printf_P(PSTR("%04x "), i);
		}
		printf_P(PSTR("\n"));
	}
	waitkey();
}


char getVlanEnabled() {
	int s=rtlRead(RTL_GENCTL);
	if (s&(0x6000)) return 1; else return 0;
}

void setVlanEnabled(char ena) {
	int s=rtlRead(RTL_GENCTL);
	if (ena) s|=0x6000; else s&=~0x6000;
	rtlWrite(RTL_GENCTL,s);
	//Hack: mapping from ports to member-id-entries. Somehow, the device doesn't set this itself.
	rtlWrite(RTL_VLANCTL_BASE,0x3210);
	rtlWrite(RTL_VLANCTL_BASE+1,0x7654);
}

void showPortSettings(int port) {
	int s=rtlRead(RTL_PORTSTATUS_BASE+(port>>1));
	if (port&1) s>>=8; else s&=0xff;
	printf_P(PSTR("Port %i: "), port+1);
	if (s&RTL_PORTSTATUS_LINKMASK) {
		printf_P(PSTR("link up: "));
		if ((s&RTL_PORTSTATUS_SPEEDMASK)==0) printf_P(PSTR("10MBit "));
		if ((s&RTL_PORTSTATUS_SPEEDMASK)==1) printf_P(PSTR("100MBit "));
		if ((s&RTL_PORTSTATUS_SPEEDMASK)==2) printf_P(PSTR("1GBit "));
		if (s&RTL_PORTSTATUS_DUPLEXMASK) printf_P(PSTR("HD")); else printf_P(PSTR("FD"));
	} else {
		 printf_P(PSTR("link down"));
	}
	printf_P(PSTR(".\n"));
}


void dumpAvrEeprom(void) {
	int x,y;
	for (y=0; y<0x100; y+=8) {
//		if (y==0x80) y=0x100;
		printf_P(PSTR("%03x: "), y);
		for (x=0; x<8; x++) {
			unsigned int i=eeprom_read_word((y+x)*2);
			printf_P(PSTR("%04x "), i);
		}
		printf_P(PSTR("\n"));
	}
	waitkey();
}

//Save the config to the internal EEPROM.
void saveConfig(void) {
	unsigned int* p;
	p=0;
	printf_P(PSTR("Saving, this takes a while...\n"));
	eeprom_write_word(p++, (0xA5<<8)+getVlanEnabled());
	for (int x=0; x<8; x++) {
		eeprom_write_word(p++, getPortVid(x));
	}
	for (unsigned int vid=0; vid<4096 && (int)p<511; vid++) {
		unsigned int r1, r2, r3;
		rtlWrite(RTL_VLANTABLE_WRITEBASE, vid);
		rtlWrite(RTL_VLANTABLE_ACCESSCTL, RTL_VLANTABLE_DO_READ);
		r1=rtlRead(RTL_VLANTABLE_READBASE);
		r2=rtlRead(RTL_VLANTABLE_READBASE+1);
		r3=rtlRead(RTL_VLANTABLE_READBASE+2);
		if (r2!=0) {
			eeprom_write_word(p++, vid);
			eeprom_write_word(p++, r2);
		}
	}
	eeprom_write_word(p++, 0xffff);
	if ((int)p>=511) {
		printf_P(PSTR("Couldn't save config: too many vlans!\n"));
	} else {
		printf_P(PSTR("Config saved; %i bytes used.\n"), (int)p);
	}
}

//Load the config from the internal EEPROM.
void loadConfig(void) {
	unsigned int *p;
	p=0;
	unsigned int vid, r2, n;
	n=eeprom_read_word(p++);
	if ((n>>8)!=0xA5) {
		printf_P(PSTR("Invalid config in eeprom; not loaded.\n"));
	//	return;
	}
	setVlanEnabled(n&0xff);
	for (int x=0; x<8; x++) {
		setPortVid(x, eeprom_read_word(p++));
	}
	while ((vid=eeprom_read_word(p++))!=0xffff) {
		r2=eeprom_read_word(p++);
		rtlWrite(RTL_VLANTABLE_WRITEBASE, vid);
		rtlWrite(RTL_VLANTABLE_WRITEBASE+1, r2);
		rtlWrite(RTL_VLANTABLE_WRITEBASE+2, 0);
		rtlWrite(RTL_VLANTABLE_ACCESSCTL, RTL_VLANTABLE_DO_WRITE);
	}
}


void resetVlans(void) {
	//Reset chip and wait...
	rtlWrite(0x100,1);
	_delay_ms(1000); //needs 10 secs?
	setVlanEnabled(1);
	for (int x=0; x<8; x++) setPortVid(x,1);
	rtlWrite(RTL_VLANTABLE_WRITEBASE, 1);
	rtlWrite(RTL_VLANTABLE_WRITEBASE+1, 0xffff);
	rtlWrite(RTL_VLANTABLE_WRITEBASE+2, 0);
	rtlWrite(RTL_VLANTABLE_ACCESSCTL, RTL_VLANTABLE_DO_WRITE);
}


char mainMenu[] PROGMEM="Main menu\n" \
		"General Settings\n" \
		"Per-port Settings\n" \
		"Info\n" \
		"Debug\n" \
		"Save config\n" \
		"Load config";

char generalMenu[] PROGMEM="General menu\n" \
		"Toggle VLAN support (is %)\n"
		"Reset to sensible VLAN settings";

void generalMenuCallback(int menuItem, int no, char *buff) {
	if (getVlanEnabled()) strcpy(buff,"Enabled"); else strcpy(buff,"Disabled");
}

char debugMenu[] PROGMEM="Debug menu\n" \
		"Dump 24c02\n" \
		"Dump RTL registers\n" \
		"Reset chip\n" \
		"Write to 24c02\n" \
		"Set register\n" \
		"Dump AVR EEPROM";

char infoMenu[] PROGMEM="Info\n" \
		"Port connections\n" \
		"VLAN config";


char portMenu[] PROGMEM="Port settings for port %\n" \
		"Connection info\n" \
		"Add to vlan, untagged output\n" \
		"Add to vlan, tagged output\n" \
		"Remove from vlan\n" \
		"Change VLAN untagged packets go to (is %)\n" \
		"View current vlan setup";

void portMenuCallback(int menuItem, int no, char *buff) {
	if (menuItem==0) sprintf(buff, "%i", currentPort+1);
	if (menuItem==5) sprintf(buff, "%i", getPortVid(currentPort));
}


//Main routine
int main(void) {
	int choice;
	stdoutInit(18); //19200@6MHz
	initIo();
	for (int x=0; x<12; x++) {
		_delay_ms(500);
		putchar(".");
	}
	loadConfig();
	while(1) {
		choice=menuShow(mainMenu, 0);
		if (choice==1) {
			//General menu
			while((choice=menuShow(generalMenu, generalMenuCallback))!=0) {
				if (choice==1) {
					setVlanEnabled(!getVlanEnabled());
				} else if (choice==2) {
					printf_P(PSTR("This will reset your current config. You sure? (y/n)\n"));
					if (getchar()=='y') {
						resetVlans();
					}
				}
			}
		} else if (choice==2) {
			//Per-port menu
			int port=0;
			printf_P(PSTR("Which port? (1-5)"));
			port=getdec()-1;
			currentPort=port;
			while(port<5 && (choice=menuShow(portMenu, portMenuCallback))!=0) {
				if (choice==1) {
					showPortSettings(port);
					waitkey();
				} else if (choice==2 || choice==3 || choice==4) {
					printf_P(PSTR("Which VID? "));
					int vid=getdec();
					if (choice==2) {
						setVlanMembership(port, VLAN_UNTAGGED, vid);
						printf_P(PSTR("Port added to VLAN, outputs untagged data.\n"));
					} else if (choice==3) {
						setVlanMembership(port, VLAN_TAGGED, vid);
						printf_P(PSTR("Port added to VLAN, outputs tagged data.\n"));
					} else if (choice==4) {
						setVlanMembership(port, VLAN_NONE, vid);
						printf_P(PSTR("Port removed.\n"));
					}
					waitkey();
				} else if (choice==5) {
					printf_P(PSTR("Tag with which VID? "));
					int vid=getdec();
					setPortVid(port, vid);
				} else if (choice==6) {
					dumpVlanInfo();
					waitkey();
				}
			}
		} else if (choice==3) {
			//Info menu
			while((choice=menuShow(infoMenu, 0))!=0) {
				if (choice==1) {
					for (int x=0; x<5; x++) showPortSettings(x);
					waitkey();
				} else if (choice==2) {
					if (!getVlanEnabled()) {
						printf_P(PSTR("Vlans are disabled! Enable them first.\n"));
					} else {
						dumpVlanInfo();
					}
					waitkey();
				}
			}
		} else if (choice==4) {
			//debug menu
			while((choice=menuShow(debugMenu, 0))!=0) {
				if (choice==1) {
					dumpEeprom();
				} else if (choice==2) {
					dumpRtlRegisters();
				} else if (choice==3) {
					rtlWrite(0x100, 1);
				} else if (choice==4 || choice==5) {
					int addr, data=0;
					printf_P(PSTR("Address?"));
					addr=gethex();
					do {
						printf_P(PSTR("%04x? "),addr);
						data=gethex();
						if (choice==4 && data!=-1) eepromWrite(addr, data);
						if (choice==5 && data!=-1) rtlWrite(addr, data);
						addr++;
					} while (data!=-1);
				} else if (choice==6) {
					dumpAvrEeprom();
				}
			}
		} else if (choice==5) {
			saveConfig();
			waitkey();
		} else if (choice==6) {
			loadConfig();
			waitkey();
		}
	}
}

